// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtkDebugRangeIterators_h
#define vtkDebugRangeIterators_h

/* Debug for-range iterator support.  */
/* #undef VTK_DEBUG_RANGE_ITERATORS */

/* Debug for-range iterator support.  */
/* #undef VTK_ALWAYS_OPTIMIZE_ARRAY_ITERATORS */

#endif
