
#ifndef VTKRENDERINGGL2PSOPENGL2_EXPORT_H
#define VTKRENDERINGGL2PSOPENGL2_EXPORT_H

#ifdef VTKRENDERINGGL2PSOPENGL2_STATIC_DEFINE
#  define VTKRENDERINGGL2PSOPENGL2_EXPORT
#  define VTKRENDERINGGL2PSOPENGL2_NO_EXPORT
#else
#  ifndef VTKRENDERINGGL2PSOPENGL2_EXPORT
#    ifdef RenderingGL2PSOpenGL2_EXPORTS
        /* We are building this library */
#      define VTKRENDERINGGL2PSOPENGL2_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define VTKRENDERINGGL2PSOPENGL2_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef VTKRENDERINGGL2PSOPENGL2_NO_EXPORT
#    define VTKRENDERINGGL2PSOPENGL2_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef VTKRENDERINGGL2PSOPENGL2_DEPRECATED
#  define VTKRENDERINGGL2PSOPENGL2_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef VTKRENDERINGGL2PSOPENGL2_DEPRECATED_EXPORT
#  define VTKRENDERINGGL2PSOPENGL2_DEPRECATED_EXPORT VTKRENDERINGGL2PSOPENGL2_EXPORT VTKRENDERINGGL2PSOPENGL2_DEPRECATED
#endif

#ifndef VTKRENDERINGGL2PSOPENGL2_DEPRECATED_NO_EXPORT
#  define VTKRENDERINGGL2PSOPENGL2_DEPRECATED_NO_EXPORT VTKRENDERINGGL2PSOPENGL2_NO_EXPORT VTKRENDERINGGL2PSOPENGL2_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef VTKRENDERINGGL2PSOPENGL2_NO_DEPRECATED
#    define VTKRENDERINGGL2PSOPENGL2_NO_DEPRECATED
#  endif
#endif

/* VTK-HeaderTest-Exclude: vtkRenderingGL2PSOpenGL2Module.h */

/* Include ABI Namespace */
#include "vtkABINamespace.h"
/* AutoInit dependencies. */
#include "vtkRenderingOpenGL2Module.h"


/* AutoInit implementations. */
#ifdef vtkRenderingGL2PSOpenGL2_AUTOINIT_INCLUDE
#include vtkRenderingGL2PSOpenGL2_AUTOINIT_INCLUDE
#endif
#ifdef vtkRenderingGL2PSOpenGL2_AUTOINIT
#include "vtkAutoInit.h"
VTK_MODULE_AUTOINIT(vtkRenderingGL2PSOpenGL2)
#endif

#endif /* VTKRENDERINGGL2PSOPENGL2_EXPORT_H */
